/*****************************************************************************/
/*                                                                           */
/*    Device driver for CC7700/PCI on Linux                                  */
/*                                                                           */
/*               Modiied  for CC7700/PCI by Naohito Iwasa on 3 Dec 1997      */
/*                                                                           */
/*****************************************************************************/

#define LINUX_2_4

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
/*#include <linux/malloc.h>*/
#include <linux/ioctl.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/config.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/poll.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>



#include "camac.h"

#define VENDOR_ID 1
#define DEVICE_ID 0xcc77
#define PORT_SIZE 0x10

static unsigned short BASE_ADDRESS;
static char IRQ;

#define HARDWARE "CC7700/PCI"
static int    dc_int_flag = 0, dc_int_lost_flag=0;
/*static struct wait_queue *wait_que=NULL;*/
static DECLARE_WAIT_QUEUE_HEAD(wait_que);

static int    wakeups     = 0;

#define EV_MAX 65536
static unsigned int   dc_int_count, dc_int_lost;
static unsigned int   dc_rp=0,      dc_wp=0;
static unsigned int   dc_ev_max=EV_MAX;
static unsigned short      dc_evbuff[EV_MAX];
static unsigned short      cc_buffer[2048];
static int shmid;


static int  dc_open   ();
static int dc_close  ();
static ssize_t  dc_read   ();
/*static ssize_t  dc_write  ();*/
/*static int  dc_select ();*/
static unsigned int  dc_poll ();
static void dc_interrupt();
static int  dc_ioctl  ();

static void dc_int_init();
static void dc_int_start();
static void dc_int_stop();
static void dc_int_clear();

static struct file_operations dc_fops = {
  THIS_MODULE /*module owner*/,
  NULL	/* dc_lseek */,
  dc_read,
  NULL  /* write   */,
  NULL	/* readdir */,
  dc_poll,
  dc_ioctl,
  NULL	/* mmap */,
  dc_open,
  NULL  /* flush */,
  dc_close,
  NULL /* fsync */,
  NULL /* fasync */,
  NULL /* check media change */,
  NULL /* revalidate */,
  NULL /* lock */
};	/* if not defined, replace to NULL. */


static int cycle_wait( int stimes ) {
  register int i;
  for(i=0;i<stimes;i++) if((inl(BASE_ADDRESS)&0x10)!=0) return 0;
  return -1;
}

static int  dc_open   (struct inode *inode, struct file *file) { 
/*  printk(DEVICE ": CC7700/PCI  opened!!\n");*/
  return 0; 
}

static int dc_close  (struct inode *inode, struct file *file) 
{ 
/*  printk(DEVICE ":CC7700/PCI closed!!");*/
  return 0;
}

static int  dc_read   ( struct file *file, 
			char *buffer, size_t count,loff_t *ppos) 
{ 
  static int i,j,nbytes=0; 
  if (dc_wp<=dc_rp) {
    wakeups++;
    interruptible_sleep_on(&wait_que);
    if(wakeups) wakeups--;
    /*    if(current->signal.sig[0] & ~current->blocked.sig[0]) 
	  return -ERESTARTNOINTR;*/
    if (signal_pending(current)) return -ERESTARTNOINTR;

  }
  nbytes=2*(dc_wp-dc_rp);
  if(nbytes>count) nbytes=((count)>>1)<<1;
  if(2*(dc_rp%EV_MAX)+nbytes>=2*EV_MAX) {
    copy_to_user(buffer,
		 &dc_evbuff[dc_rp%EV_MAX],
		 2*(EV_MAX-dc_rp%EV_MAX));
    copy_to_user(&buffer[2*(EV_MAX-dc_rp%EV_MAX)],
		 dc_evbuff,
		 nbytes-2*(EV_MAX-dc_rp%EV_MAX));
  } 
  else {
    copy_to_user(buffer,
		 &dc_evbuff[dc_rp%EV_MAX],
                 nbytes);
  }
  dc_rp+=nbytes/2;
  if(dc_int_lost_flag!=0&& (dc_wp-dc_rp +cc_buffer[0]) <= EV_MAX-1024) {
    dc_evbuff[(dc_wp++)%EV_MAX]=DELIM;
    dc_evbuff[(dc_wp++)%EV_MAX]=j=cc_buffer[0];
    for(i=1;i<j;i++){
      dc_evbuff[(dc_wp++)%EV_MAX]=cc_buffer[i];
    }
    dc_int_lost_flag=0;
    dc_int_lost--;
    dc_int_clear();
    /*    printk("lost recovered %d\n",dc_int_flag);*/
    if(dc_int_flag!=0)
      while((inl(BASE_ADDRESS)&0x08)==0)
	outl(0x0008,BASE_ADDRESS);
    
  }
  return nbytes;
}


static int dc_write (struct file * file, 
		     char * buffer, int count, loff_t *ppos) 
{ return 0; }


static unsigned int dc_poll(struct file *file,poll_table *wait) 
{
  unsigned int mask;
  long lampat;

  mask=0;
  if(dc_wp<=dc_rp) {
    poll_wait(file,&wait_que,wait);
  }
  if(dc_wp>dc_rp){
    mask|=(POLLIN | POLLRDNORM);
  }
  lampat=inl(BASE_ADDRESS+4);
  if(lampat==0) {
    poll_wait(file,&wait_que,wait);
  } 
  else {
    mask |= (POLLPRI);
  }
  return mask;
}

#include "dc_int.c"

static void dc_interrupt(int irq, struct pt_regs *regs)
{
  static int i,j;

  dc_int_count++;

  cc_buffer[0]=0;
  i=dc_int(cc_buffer);
  if(i>=0){ 
    if ((dc_wp-dc_rp +cc_buffer[0]) > EV_MAX-1024) {  /* buffer is full */
      dc_int_lost++;
      dc_int_lost_flag=1;
      while((inl(BASE_ADDRESS)&0x08)!=0) outl(0x00,BASE_ADDRESS);
      /*      printk("data lost  %d\n",dc_int_flag);*/
    } 
    else {
      dc_evbuff[(dc_wp++)%EV_MAX]=DELIM;
      dc_evbuff[(dc_wp++)%EV_MAX]=j=cc_buffer[0];
      for(i=1;i<j;i++){
	dc_evbuff[(dc_wp++)%EV_MAX]=cc_buffer[i];
      }
      dc_int_clear();
    }
  }
  else{
    dc_int_clear();
  }
  if (wakeups) wake_up(&wait_que);
}


static int dc_ioctl(struct inode *inode, struct file *file,
		    unsigned int cmd, unsigned long arg) 
{
  static int ans=0,i;
  static long CNAF;

  cli();
  switch(cmd) {

  case DC_INITIAL:
/*    if (arg > 7) {ans=-1; break;}*/
/*    outl(arg<<16,BASE_ADDRESS+4);*/
/*    outl((inl(BASE_ADDRESS)&0x8c)|0x02,BASE_ADDRESS);*/
    if((inl(BASE_ADDRESS)&0x20)!=0) ans=0; else ans=1;
    break;

/************************
* CAMAC single function *
************************/

  case DC_Z: 
    outl((inl(BASE_ADDRESS)&0x8c)|0x02,BASE_ADDRESS);
    ans=cycle_wait(1000);
    break;
  case DC_C: 
    outl((inl(BASE_ADDRESS)&0x8c)|0x01,BASE_ADDRESS);
    ans=cycle_wait(1000);
    break;
  case DC_I:
    if(arg!=0) 
      while((inl(BASE_ADDRESS)&0x04)==0)
	    outl(0x04,BASE_ADDRESS);
    else{ 
      while((inl(BASE_ADDRESS)&0x04)!=0)
	     outl(0x00,BASE_ADDRESS);
      if(dc_int_flag!=0)
	while((inl(BASE_ADDRESS)&0x08)==0)
	  outl(0x0008,BASE_ADDRESS);
    }
    break;
  case DC_L_P: {
    put_user(inl(BASE_ADDRESS+4),(int *)arg);
    break;
  }
  case DC_L_E: {
    int i; 
    long lampat;
    lampat = inl((BASE_ADDRESS+4)&0x0000ffff);
    for(i=0;i<24;i++) if((lampat>>i)%2!=0) goto lamhit;
  lamhit:
    if(i<24) i++; else i=0;
    put_user(i, (int *)arg);
    break;
  }
  case DC_CYCLE: {
    unsigned short na;  DCCycle dcc;
    copy_from_user(&dcc,(unsigned long *)arg,sizeof(DCCycle));
    outl(((long)dcc.f)|((long)dcc.a)<<5|((long)dcc.n-1)<<9|((long)dcc.c)<<16,BASE_ADDRESS+4);
    if((dcc.f&0x18)==0x10){
      outl((long)dcc.data,BASE_ADDRESS+8);
    }
    outl(0,BASE_ADDRESS+12);
/*    printk(DEVICE ": add=%8lx\n",BASE_ADDRESS+12);*/
    ans=cycle_wait(512);
    dcc.stat=((~inl(BASE_ADDRESS))&0x03);
    if((dcc.f&0x18)==0x00) {
      dcc.data=inl((BASE_ADDRESS+8)&0x0000ffff);
/*      printk(DEVICE ": data=%x add=%x\n",dcc.data,(BASE_ADDRESS+8)&0x0000ffff);*/
    }
    copy_to_user((unsigned long *)arg,&dcc,sizeof(dcc));
    break;
  }

/********************
* interrupt control *
********************/

  case DC_INTTST: {
    struct pt_regs regs;
    dc_interrupt(IRQ,&regs);
    break;
  }
  case DC_INTON:
    dc_evbuff[(dc_wp++)%EV_MAX]=DELIM;
    dc_evbuff[(dc_wp++)%EV_MAX]=0xfffc;
    if (wakeups) wake_up(&wait_que);
/*    dc_int_count=0;*/
    request_irq(IRQ,(void *)dc_interrupt,SA_INTERRUPT,HARDWARE,NULL);
    dc_int_flag=1;
    dc_int_start();
    outl(0x08,BASE_ADDRESS);
    break;

  case DC_INTOFF:
    dc_evbuff[(dc_wp++)%EV_MAX]=DELIM;
    dc_evbuff[(dc_wp++)%EV_MAX]=0xfffd;
    if (wakeups) wake_up(&wait_que);
    outl(0x00,BASE_ADDRESS);
    dc_int_stop();
    if(dc_int_flag!=0)free_irq(IRQ,NULL);
    dc_int_flag=0;
    break;

  case DC_INTCLR:
    dc_evbuff[0]=DELIM;
    dc_evbuff[1]=0xfffb;
    dc_rp=0;
    dc_wp=2;
    dc_int_lost=0;
    dc_int_count=0;
    if (wakeups) wake_up(&wait_que);
    break;

  case DC_INTCNT:
    put_user(dc_int_count, (int *) arg);
    break;

  case DC_INTLOS:
    put_user(dc_int_lost,(int *) arg);
    break;

  case DC_INTBUF:
    put_user(dc_wp-dc_rp,(int *) arg);
    break;

  case DC_INTOPN:
    dc_evbuff[0]=DELIM;
    dc_evbuff[1]=0xfffe;
    dc_rp=0;
    dc_wp=2;
    dc_int_lost=0;
    dc_int_count=0;
    if (wakeups) wake_up(&wait_que);
    break;

  case DC_INTEOF:
    dc_evbuff[(dc_wp++)%EV_MAX]=DELIM;
    dc_evbuff[(dc_wp++)%EV_MAX]=0xffff;
    if (wakeups) wake_up(&wait_que);
    break;

/*********************
* direct I/O actions *
*********************/

  case DC_WRITE_PORT1:
    outl(arg,BASE_ADDRESS);
    break;

  case DC_WRITE_PORT2:
    outl(arg,BASE_ADDRESS+4);
    break;

  case DC_WRITE_PORT3:
    outl(arg,BASE_ADDRESS+8);
    break;

  case DC_WRITE_PORT4:
    outl(arg,BASE_ADDRESS+12);
    break;

  case DC_READ_PORT1:
    ans=inl(BASE_ADDRESS);
    break;

  case DC_READ_PORT2:
    ans=inl(BASE_ADDRESS+4);
    break;

  case DC_READ_PORT3:
    ans=inl(BASE_ADDRESS+8);
    break;

/************************
* for performance check *
************************/

/*
  case DC_TEST1:
    for(i=0;i<10000000;i++){
      outl(0xffffff,BASE_ADDRESS+8);
      CNAF=17|4<<9;
      outl(CNAF,BASE_ADDRESS+4);
      outl(0,BASE_ADDRESS+12);
      cycle_wait(255);
    }
    break;

  case DC_TEST2:
    CNAF=17|4<<9;
    for(i=0;i<10000000;i++){
      outl(0xffffff,BASE_ADDRESS+8);
      outl(CNAF,BASE_ADDRESS+4);
      outl(0,BASE_ADDRESS+12);
      cycle_wait(255);
    }
    break;

  case DC_TEST3:
    CNAF=17|4<<9;
    outl(CNAF,BASE_ADDRESS+4);
    for(i=0;i<10000000;i++){
      outl(0xffffff,BASE_ADDRESS+8);
      outl(0,BASE_ADDRESS+12);
      cycle_wait(255);
    }
*/
    break;
  default: 
    break;
  }
  sti();
  return ans;
}

#ifndef MODULE
unsigned long dc_init(unsigned long kmem_start) 
{
#else
int init_module(void) 
{
#endif

  unsigned char pci_bus,pci_device_fn;
  int i;

  printk("-------------------------<<< PCIcc7700 >>>--------------------------.\n");

  if(!pcibios_present()){
    printk(DEVICE ": PCI bios does not exist.\n");
#ifndef MODULE
    return kmem_start;
#else
    return -ENODEV;
#endif
  }
  if(pcibios_find_device(VENDOR_ID,DEVICE_ID,0,
                      &pci_bus,&pci_device_fn)){
    printk(DEVICE ": A PCI board for " DEVICE " was not found\n");
#ifndef MODULE
    return kmem_start;
#else
    return -EIO;
#endif
  }
  if(PNO>0) BASE_ADDRESS=PNO & 0xfffc;
  else {
    pcibios_read_config_dword(pci_bus, pci_device_fn, PCI_BASE_ADDRESS_0,
                             &i);
    BASE_ADDRESS=i & 0xfffc;
  }

  for(i=0;i<5;i++){
    if(!check_region(BASE_ADDRESS,PORT_SIZE))break;
    BASE_ADDRESS+=((PORT_SIZE-1)/0x04+1)*0x04;
  }
  if(i==5){
    printk(DEVICE ": unable to allocate I/O addresses from %08x to %08x.\n",
           PNO & 0xfffc, BASE_ADDRESS);
#ifndef MODULE
    return kmem_start;
#else
    return -ENODEV;
#endif
  }

  printk(DEVICE ":allocating I/O address %x: %d\n",BASE_ADDRESS,
  pcibios_write_config_dword(pci_bus, pci_device_fn, PCI_BASE_ADDRESS_0,
                             0x0000ffff & BASE_ADDRESS));


   pcibios_read_config_byte( pci_bus, pci_device_fn, PCI_INTERRUPT_LINE,
      &IRQ);
 /*
  printk(DEVICE "allocated IRQ address %d\n",c);

  printk(DEVICE "allocating IRQ address %d\n",
  pcibios_write_config_byte( pci_bus, pci_device_fn, PCI_INTERRUPT_LINE,
                             0xff & IRQ));
  pcibios_read_config_byte( pci_bus, pci_device_fn, PCI_INTERRUPT_LINE,
                             &c);
  printk(DEVICE "allocated IRQ address %d\n",c);
*/

  request_region(BASE_ADDRESS,PORT_SIZE,DEVICE);

  outl(0x00,BASE_ADDRESS);

  if((i=request_irq(IRQ,(void *)dc_interrupt,SA_INTERRUPT,HARDWARE,NULL))){
    printk(DEVICE ": IRQ=%02x was occupied.\n",IRQ);
    release_region(BASE_ADDRESS,PORT_SIZE);
#ifndef MODULE
    return kmem_start;
#else
    return(i);
#endif
  }

  if(register_chrdev(DC_NO,DEVICE,&dc_fops)){
    printk(DEVICE ": Major number %d was occupied.\n",DC_NO);
    free_irq(IRQ,NULL);
    release_region(BASE_ADDRESS,PORT_SIZE);
#ifndef MODULE
    return kmem_start;
#else
    return -EIO;
#endif
  }
  printk(DEVICE ":" DEVICE " was installed.    BASE=0x%04x IRQ=0x%02x\n",
         BASE_ADDRESS,IRQ);
  printk(DEVICE ": Buffer size is %d bytes\n", 2 * EV_MAX);

/*
  setCSR(CC_CSRM_RST);
*/

/*  if((shmid = sys_shmget(1048577, 65536, IPC_CREAT|0600))==-1){
    printk(DEVICE "shared memory allocation failed\n");
  }
  else{
    printk(DEVICE "shared memory allocation failed\n");
  }
*/
#ifndef MODULE
  return kmem_start;
#else
  return 0;
#endif
}

#ifndef MODULE
static void cleanup_module(void) {
#else
void cleanup_module(void) {
#endif

  printk("-------------------------<<< PCIcc7700 >>>--------------------------.\n");
  if (wakeups) wake_up(&wait_que);  
  free_irq(IRQ,NULL); 
  release_region(BASE_ADDRESS,PORT_SIZE);
  unregister_chrdev(DC_NO,DEVICE);
  printk(DEVICE ":" DEVICE " was uninstalled.\n");
}
